import json
import logging
import os
import zlib

import bson
import click

from antares_client import Client

logger = logging.getLogger("cli")
logger.setLevel(logging.DEBUG)


@click.command()
@click.option("-v", "--verbose")
@click.option("-h", "--host", help="Hostname of Kafka cluster", type=str)
@click.option("-p", "--port", help="Port of Kafka cluster", type=int)
@click.option(
    "-g",
    "--group",
    help="Globally unique name of consumer group (default, hostname)",
    type=str,
)
@click.option(
    "-d",
    "--output-directory",
    type=str,
    help="Directory to save alerts in. If not passed, a message will be logged to STDOUT when alerts are received.",
)
@click.option("--api-key", help="ANTARES Kafka API Key", type=str)
@click.option("--api-secret", help="ANTARES Kafka API Secret", type=str)
@click.option(
    "--ssl-ca-location",
    help="Location of your SSL root CAs cert.pem file",
    type=click.Path(exists=True),
)
@click.argument("topics", nargs=-1, required=True)
def stream(**kwargs):
    """
    Stream alerts from ANTARES

    This command subscribes to ANTARES streams and writes alerts to ``*.json``
    files in a directory of your choice.

    For example, to save all alerts from the ``extragalactic`` stream to a
    folder in your home directory called ``antares/extragalactic``::

    \b
    antares stream extragalactic
        --output-directory ~/antares
        --api-key ********************
        --api-secret ********************

    The client creates a subdirectory in ``--output-directory`` named after the
    stream you're subscribing to and saves that stream's alerts there (in this
    case, ``~/antares/extragalactic``).

    You can also subscribe to multiple streams::

    \b
    antares stream extragalactic nuclear_transient
      --output-directory ~/antares
      --api-key ********************
      --api-secret ********************

    And this will save the alerts from the ``extragalactic`` and
    ``nuclear_transient`` streams in ``~/antares/extragalactic`` and
    ``~/antares/nuclear_transient``, respectively.

    """
    logger.info("Opening connection to ANTARES Kafka...")
    kwargs["topics"] = list(kwargs["topics"])
    if len(kwargs["topics"]) == 1:
        kwargs["topics"] = kwargs["topics"][0].split(",")
        if len(kwargs["topics"]) > 1:
            logger.warning(
                "comma-separated topics have been deprecated in favor of space-separated topics"
            )
    with Client(**kwargs) as client:
        for topic, alert in client.iter():
            if kwargs.get("output_directory", False):
                directory = os.path.join(kwargs["output_directory"], topic)
                if not os.path.exists(directory):
                    os.makedirs(directory)
                alert_id = get_alert_id(alert)
                file_name = "{}.json".format(alert_id)
                file_path = os.path.join(directory, file_name)
                with open(file_path, "w") as f:
                    json.dump(alert, f, indent=4)
                logger.info("Saved alert {}".format(file_path))
            else:
                logger.info("Received alert on topic '{}'".format(topic))
        logger.info("Closing connection...")


def get_alert_id(alert):
    alert_id = alert["new_alert"].get("alert_id", None)
    if alert_id is None:
        alert_id = "{}-{}".format(
            alert["new_alert"]["survey"], alert["new_alert"]["original_id"]
        )
    return alert_id


def parse_antares_alert(payload):
    """
    Convert an ANTARES Alert message to a Python object.

    ANTARES Alerts are outputted in GZIP-compressed BSON format.

    :param payload: byte array of message
    :return: dict
    """
    try:
        return bson.loads(zlib.decompress(payload))
    except Exception:
        logger.error("Failed to parse message:")
        logger.error(payload)
        raise



import logging


def setup_logger(name):
    formatter = logging.Formatter(
        fmt="%(asctime)s - %(levelname)s - %(module)s - %(message)s"
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return logger



import json
import logging
import sys

import click

import antares_client
from antares_client.search import JobStatus

logger = logging.getLogger("cli")
logger.setLevel(logging.DEBUG)


def search_progress_callback(status, **kwargs):
    if status == JobStatus.SUBMITTED:
        logger.info("Job {id} - Submitted to ANTARES server".format(**kwargs["job"]))
        logger.info("Job {id} - Status URL: {urls[status]}".format(**kwargs["job"]))

    elif status == JobStatus.PROCESSING:
        logger.info(
            "Job {id} - ANTARES server processed {recordsProcessed} / {recordsTotal} records".format(
                **kwargs["job"]
            )
        )

    elif status == JobStatus.READY:
        logger.info("Job {id} - ANTARES finished processing".format(**kwargs["job"]))
        logger.info("Job {id} - Download URL: {urls[download]}".format(**kwargs["job"]))

    elif status == JobStatus.DOWNLOADING:
        logger.info(
            "Job {id} - Downloading results from ANTARES server".format(**kwargs["job"])
        )

    elif status == JobStatus.COMPLETED_DOWNLOADING:
        logger.info("Job {id} - Completed download".format(**kwargs["job"]))

    elif status == JobStatus.ERRORED:
        logger.error("Job {id} - Fatal error".format(**kwargs["job"]))


@click.command()
@click.option(
    "--output-format",
    type=click.Choice(["json", "csv"]),
    default="json",
    help="Format of output file (one of: ``{'json', 'csv'}``, default, ``'json'``)",
)
@click.option(
    "--decompress/--compress",
    default=False,
    help="Decompress or compress (gzip) the results before writing to file (default, ``compress``)",
)
@click.option(
    "-o",
    "--output",
    type=click.File("wb"),
    default="-",
    help="Location to write output file (default, ``STDOUT``)",
)
@click.argument("query-file", type=click.File("rb"))
def search(query_file, output, decompress, output_format):
    """
    Download an ElasticSearch query result set from ANTARES.

    If you have a file ``query.json`` with an ElasticSearch query, e.g.::

    \b
    {
      "query": {
        "range": {
          "locus_id": {
            "gte": 100,
            "lte": 20000
          }
        }
      }
    }

    Then the command:

    ``antares search -o output.json query.json``

    Will download the results of the search query to ``output.json``.

    These commands play nicely with Unix standard streams so this is
    equivalent:

    ``cat query.json | antares-client search -o output.json -``

    To write the results to STDOUT, do not pass anything for the
    ``-o/--output`` option:

    ``cat query.json | antares-client search -``

    """
    query = json.loads(query_file.read().decode("utf-8"))
    logger.info("Submitting query to ANTARES server: {}".format(query))
    try:
        result_set = antares_client.search._raw_search(  # pylint: disable=protected-access
            query,
            output_format=output_format,
            decompress=decompress,
            progress_callback=search_progress_callback,
        )
        output.write(result_set)
    except:  # pylint: disable=bare-except
        logger.exception("Encountered a fatal error:")
        sys.exit(1)



"""
Connect to ANTARES' Kafka cluster and read messages.

"""
import click

import antares_client
from . import log
from .commands.search import search
from .commands.stream import stream

logger = log.setup_logger("cli")


@click.group()
@click.version_option(antares_client.__version__)
def entry_point():
    pass


entry_point.add_command(search)
entry_point.add_command(stream)



"""
This module wraps the ElasticSearch query functionality provided by
the ANTARES portal.

Downloading an ES result-set works as follows:

    1) The client POSTs to ``/alerts/search/download`` with the ES
       query;

    2) The server submits the query to another thread for processing.
       The server stores the name of this thread, along with the job's
       status, progress, etc., in Redis. The server responds to the
       client with a JSON representation of this ``job`` object.

    3) The server creates a file in a temporary directory named
       ``{job.id}.results.csv.gz``;

    4) The server incrementally queries for paged results from the
       ESDB and writes them into ``{job.id}.results.csv.gz``.

Any point after step (2), the client can GET the
``/alerts/search/downloads/<str:job_id>/status`` route to check the
status of the job.

The response from the server has the following format:

.. code:: json

   {
     "job": {
       "id": UUID4,
       "status": "submited" | "processing" | "ready" | "errored"
       "format": "json" | "csv",
       "created_at": <str>,
       "updated_at": <str>,
       "recordsTotal": <int>,
       "recordsProcessed": <int>,
       "urls": {
         "status": <str>,
         "download": <str>
       }
     }
   }

Once ``job["status"] == "ready"`` the client may GET the
``/alerts/search/download/<str:job_id>`` route to download the result
set.

"""

import enum
import gzip
import json
import time

import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry


class SearchError(Exception):
    """Error evaluating ElasticSearch query"""


class JobStatus(enum.Enum):
    """Status messages for progress callbacks"""

    SUBMITTED = "submitted"  #: Job has been submitted to the server
    PROCESSING = (
        "processing"
    )  #: Job has been received by the server and is being processed
    READY = "ready"  #: Job has been finished processing by the server
    DOWNLOADING = "downloading"  #: Job is being downloaded from the server
    COMPLETED_DOWNLOADING = (
        "completed_downloading"
    )  #: Job is finished downloading from the server
    ERRORED = "errored"  #: An error occured either client- or server-side


BASE_URL = "https://antares.noao.edu"
NOOP = (
    lambda *args, **kwargs: None
)  #: Noop function that accepts arbitrary ``args``, ``kwargs``
REQUESTS_RETRIES = 2
REQUESTS_BACKOFF_FACTOR = 0.1
REQUESTS_STATUS_FORCELIST = (500, 502, 504)

REQUESTS_POLL_FREQUENCY = 3


def configure_requests_session(
    session=None,
    retries=REQUESTS_RETRIES,
    backoff_factor=REQUESTS_BACKOFF_FACTOR,
    status_forcelist=REQUESTS_STATUS_FORCELIST,
):
    """
    Configure or create a ``requests.session`` to use an intelligent
    retry strategy.

    Notes
    ----------
    Peter Bengtsson
    https://www.peterbe.com/plog/best-practice-with-retries-with-requests

    """
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


SESSION = configure_requests_session()


def search(query, progress_callback=NOOP):
    """
    Retrieves an ElasticSearch query resultset from the ANTARES portal.

    Parameters
    ----------
    query: dict
        ElasticSearch query.
    progress_callback: callable
        Called periodically throughout the processing of ``query``, first
        argument will be an enum member of ``antares_client.search.JobStatus``.
        Should also accept ``*args``.  and ``**kwargs``.

    Raises
    ----------
    MaxRetryError
        If the server returns ``REQUESTS_STATUS_FORCELIST`` status codes
        more than ``REQUESTS_RETRIES`` times.
    ConnectionError
        If a connection to the server cannot be established.
    antares_client.search.SearchError
        If the server encounters an error processing the search query.

    Returns
    ----------
    result_set: [dict]

    """
    result_set = _raw_search(
        query,
        output_format="json",
        decompress=True,
        progress_callback=progress_callback,
    )
    result_set = json.loads(result_set.decode("utf-8"))
    return result_set


def download(
    query, path, output_format="json", decompress=False, progress_callback=NOOP
):
    """
    Downloads an ElasticSearch query result set from the ANTARES portal.

    Parameters
    ----------
    query: dict
        ElasticSearch query.
    path: str
        Path to write the result set to.
    output_format: {"json", "csv"}
        Output format of the result set (default, json).
    decompress: bool
        Un-gzip the result set (default, False).
    progress_callback: callable
        Called periodically throughout the processing of ``query``, first
        argument will be an enum member of ``antares_client.search.JobStatus``.
        Should also accept ``*args``.  and ``**kwargs``.

    Raises
    ----------
    MaxRetryError
        If the server returns ``REQUESTS_STATUS_FORCELIST`` status codes
        more than ``REQUESTS_RETRIES`` times.
    ConnectionError
        If a connection to the server cannot be established.
    antares_client.search.SearchError
        If the server encounters an error processing the search query.

    Returns
    ----------
    None

    """
    result_set = _raw_search(
        query,
        output_format=output_format,
        decompress=decompress,
        progress_callback=progress_callback,
    )
    with open(path, "wb") as f:
        f.write(result_set)


def _raw_search(query, output_format="json", decompress=False, progress_callback=NOOP):
    if output_format not in {"json", "csv"}:
        raise ValueError("output_format must be one of {'json', 'csv'}")
    job = _submit_search_job(query, output_format, progress_callback=progress_callback)
    job = _block_until_search_job_is_done(job, progress_callback=progress_callback)
    result_set = _download_search_result_set(
        job, decompress=decompress, progress_callback=progress_callback
    )
    return result_set


def _submit_search_job(query, output_format="json", progress_callback=NOOP):

    response = SESSION.post(
        BASE_URL + "/alerts/search/download/",
        {"es_query": json.dumps(query), "format": output_format},
    )
    response.raise_for_status()
    job = response.json()["job"]
    progress_callback(JobStatus.SUBMITTED, job=job)
    return job


def _block_until_search_job_is_done(job, timeout=float("inf"), progress_callback=NOOP):
    """
    Blocks until the server has finished preparing the result set
    for download.

    Parameters
    ----------
    timeout: float

    Raises
    ----------
    ConnectionError
    MaxRetryError
    TimeoutError
        If polling takes more than ``timeout`` seconds

    """
    start = time.time()
    while (time.time() - start) <= timeout:
        response = SESSION.get(job["urls"]["status"])
        response.raise_for_status()
        job = response.json()["job"]
        if job["status"] == "errored":
            progress_callback(JobStatus.ERRORED, job=job)
            raise SearchError
        if job["status"] == "processing":
            progress_callback(JobStatus.PROCESSING, job=job)
        elif job["status"] == "ready":
            progress_callback(JobStatus.READY, job=job)
            return job
        time.sleep(REQUESTS_POLL_FREQUENCY)
    else:  # pylint: disable=useless-else-on-loop
        raise TimeoutError


def _download_search_result_set(job, decompress=False, progress_callback=NOOP):
    """
    Downloads the query resultset (i.e. GETs the data from the server).

    Parameters
    ----------
    timeout: float

    Returns
    ----------
    bytes

    Raises
    ----------
    ConnectionError
    MaxRetryError

    """
    progress_callback(JobStatus.DOWNLOADING, job=job)
    response = SESSION.get(job["urls"]["download"])
    response.raise_for_status()
    result_set = response.content
    if decompress:
        result_set = gzip.decompress(result_set)
    progress_callback(JobStatus.COMPLETED_DOWNLOADING, job=job)
    return result_set



import os
import requests


THUMBNAIL_TYPES = ["template", "science", "difference"]
THUMBNAIL_URL = "http://antares.noao.edu/alerts/data/{alert_id}/thumbnails/{img_type}"


def get_thumbnails(alert_id, output_dir=None):
    """
    Download thumbnail images for a given ANTARES `alert_id`.

    If `output_dir` is given, save the files to that location.

    Returns a dict containing the three thumbnails, of structure:
    
    .. code:: json

       {
           'template': {
               'file_name': '...fits.gz',
               'file_bytes': '...'
           },
           'science': {
               'file_name': '...fits.gz',
               'file_bytes': '...'
           },
           'difference': {
               'file_name': '...fits.gz',
               'file_bytes': '...'
           }
       }

    Parameters
    ----------
    alert_id: int
        ANTARES Alert ID
    output_dir: str
        Output directory (optional, default None)

    Returns
    ----------
    dict

    Raises
    ----------
    requests.HTTPError
        If the server responds with a 4XX or 5XX status code

    """
    # Download image data
    results = {}
    for img_type in THUMBNAIL_TYPES:
        url = THUMBNAIL_URL.format(alert_id=alert_id, img_type=img_type)
        res = requests.get(url)
        res.raise_for_status()
        file_name = res.headers["Content-Disposition"].partition("=")[2]
        file_bytes = res.content
        results[img_type] = {"file_name": file_name, "file_bytes": file_bytes}

        # Output to file
        if output_dir:
            file_path = os.path.join(output_dir, file_name)
            with open(file_path, "wb") as f:
                f.write(file_bytes)

    return results
